from typing import Tuple, Union, List
import numpy as np
from sklearn.neighbors import KNeighborsClassifier
import pickle
from .search_interface import SearchInterface
from sentence_transformers import SentenceTransformer
import os


_KNN_FNAME = "knn.pkl"
_EMBEDDINGS_FNAME = "embeddings.pkl"
_TEXTS_FNAME = "texts.pkl"


class SearchLocalDatabase(SearchInterface):
    def __init__(self, model: SentenceTransformer, data_directory: str, top_n: int, encoder_batch_size: int) -> None:
        self.model = model
        self.top_n = top_n
        self.data_directory = data_directory
        self.encoder_batch_size = encoder_batch_size
        self._knn, self._embeddings, self._texts = self._load_data()

    def _load_pickle(self, fname: str) -> object:
        with open(fname, "rb") as src:
            return pickle.load(src)

    def _save_pickle(self, fname: str, data: object) -> None:
        with open(fname, "wb") as target:
            pickle.dump(data, target)

    def _load_data(self) -> Tuple[\
        Union[KNeighborsClassifier, None], \
        Union[np.ndarray, None], \
        Union[np.ndarray, None] \
        ]:
        os.makedirs(self.data_directory, exist_ok=True)
        knn_path = os.path.join(self.data_directory, _KNN_FNAME)
        embeddings_path = os.path.join(self.data_directory, _EMBEDDINGS_FNAME)
        texts_path = os.path.join(self.data_directory, _TEXTS_FNAME)
        paths = [self.data_directory, knn_path, embeddings_path, texts_path]
        for path in paths:
            if not os.path.exists(path):
                return None, None, None
        knn = self._load_pickle(os.path.join(self.data_directory, _KNN_FNAME))
        embeddings = self._load_pickle(os.path.join(self.data_directory, _EMBEDDINGS_FNAME))
        texts = self._load_pickle(os.path.join(self.data_directory, _TEXTS_FNAME))
        return knn, embeddings, texts

    def add_documents(self, documents: List[str]) -> None:
        documents = [document.strip() for document in documents]
        document_embeddings = self.model.encode(documents, batch_size=self.encoder_batch_size)
        if self._embeddings:
            new_embeddings = np.vstack([self._embeddings, document_embeddings])
        else:
            new_embeddings = document_embeddings
        if self._texts:
            new_texts = np.vstack([self._texts, documents])
        else:
            new_texts = documents
        knn = KNeighborsClassifier(n_neighbors=self.top_n)
        knn.fit(new_embeddings, new_texts)
        self._save_pickle(os.path.join(self.data_directory, _KNN_FNAME), knn)
        self._save_pickle(os.path.join(self.data_directory, _EMBEDDINGS_FNAME), new_embeddings)
        self._save_pickle(os.path.join(self.data_directory, _TEXTS_FNAME), new_texts)
        self._knn, self._embeddings, self._texts = self._load_data()

    def search(self, query: str) -> List[str]:
        if not self._knn:
            return []
        embeddings = self.model.encode([query], batch_size=self.encoder_batch_size)
        scores = self._knn.predict_proba(embeddings)[0]
        top_indices = (-scores).argsort()[:self.top_n]
        return [
            self._knn.classes_[index]
            for index in top_indices
        ]



class DialogueList(list):
    def __init__(self, keep_top_n: int, *args, **kwargs):
        super(DialogueList, self).__init__(*args, **kwargs)
        self.keep_top_n = keep_top_n
    
    def append(self, item: str) -> None:
        super(DialogueList, self).append(item)
        while len(self) > self.keep_top_n:
            del self[0]

    def __str__(self) -> str:
        result = ""
        for item in self:
            result += f"\n- {item}"
        return result.strip()



from dataclasses import dataclass
from typing import List, Tuple, Union
from .search import SearchInterface
from .completion_reactions import CompletionReaction
from .language_model import LanguageModelInterface, Completion
from .hints_dict import HintsDict
from .dialogue_list import DialogueList
import os


@dataclass
class CharacterConfig:
    name: str
    description: str


@dataclass
class RobotConfig:
    character: CharacterConfig
    search_system_local: SearchInterface
    language_model: LanguageModelInterface
    completions: List[CompletionReaction]
    hints_keep_top_n: int
    dialogue_keeps_top_n: int
    max_repeats: int = 5
    prompt_fname: str = os.path.join(os.path.dirname(__file__), "prompt.txt")


@dataclass
class Session:
    dialogue: DialogueList
    hints: HintsDict


class Robot:
    def __init__(self, config: RobotConfig) -> None:
        self.config = config
        with open(self.config.prompt_fname, "r", encoding="utf-8") as src:
            self.prompt = src.read()

    def init_session(self) -> Session:
        return Session(
            DialogueList(self.config.dialogue_keeps_top_n),
            HintsDict(self.config.hints_keep_top_n),
        )

    def get_prompt(self, hints: HintsDict, dialogue: DialogueList, phrase: str) -> str:
        filler_variables = {
            "CHARACTER_NAME": self.config.character.name,
            "CHARACTER_DESCRIPTION": self.config.character.description,
            "HINTS": str(hints),
            "DIALOGUE": str(dialogue),
            "RETORT": phrase,
        }
        prompt = self.prompt
        for name, value in filler_variables.items():
            prompt = prompt.replace(f"%{name}%", value)
        return prompt

    def generate_without_forcing_prefixes(self, prompt: str) -> Tuple[Union[Completion, None], Union[CompletionReaction, None], bool]:
        lm_completions = self.config.language_model.complete(
            prompt.strip()
        )
        if not lm_completions:
            return None, None, True
        top_completion = sorted(lm_completions, key=lambda completion: -completion.score)[0]
        continue_with_forced_prefixes = True
        top_reaction = None
        for reaction in self.config.completions:
            prefix = reaction.get_prefix()
            if top_completion.text.startswith(prefix):
                continue_with_forced_prefixes = False
                top_reaction = reaction
                top_completion = Completion(
                    text=top_completion.text.replace(prefix, "").strip(),
                    score=top_completion.score,
                )
        if continue_with_forced_prefixes:
            return None, None, True
        return top_completion, top_reaction, False

    def generate_with_forcing_prefixes(self, prompt: str) -> Tuple[Completion, CompletionReaction, bool]:
        lm_completions = []
        for reaction in self.config.completions:
            prefix_completions += self.config.language_model.complete(
                prompt + reaction.get_prefix()
            )
            lm_completions += [(item, reaction) for item in prefix_completions]
        lm_completions = sorted(
            lm_completions,
            key=lambda item: -item[0].score
        )
        top_completion, top_reaction = lm_completions[0]
        return top_completion, top_reaction, False

    def response(self, session: Session, phrase: str) -> Tuple[Union[str, None], Session]:
        hints = session.hints
        dialogue = session.dialogue
        force_prefixes = False
        for _ in range(self.config.max_repeats):
            prompt = self.get_prompt(hints, dialogue, phrase.strip())
            if not force_prefixes:
                top_completion, top_reaction, force_prefixes = self.generate_without_forcing_prefixes(prompt)
            else:
                top_completion, top_reaction, force_prefixes = self.generate_with_forcing_prefixes(prompt)
            top_completion_text = top_completion.text
            top_response = top_reaction.process(top_completion_text, hints)
            if not top_response.repeat:
                break
        new_documents = [phrase]
        if top_response.text:
            new_documents += [top_response.text]
        self.config.search_system_local.add_documents(new_documents)
        dialogue.append(phrase)
        if top_response.text:
            dialogue.append(top_response.text)
        session = Session(dialogue, hints)
        return top_response.text, session



from typing import List, Union, Callable
from collections import OrderedDict


HintPostprocessor = Callable[[str], str]


class HintsDict(OrderedDict):
    def __init__(self, keep_top_n: int, postprocess: Union[HintPostprocessor, None] = None, *args, **kwargs):
        super(HintsDict, self).__init__(*args, **kwargs)
        self.keep_top_n = keep_top_n
        if postprocess is None:
            postprocess = lambda text: text
        self.postprocess = postprocess

    def __setitem__(self, key: str, value: List[str]) -> None:
        super(HintsDict, self).__setitem__(key, value)
        if len(self.keys()) > self.keep_top_n:
            keys = self.keys()
            keys_to_remove = keys[:len(keys) - self.keep_top_n]
            for key in keys_to_remove:
                del self[key]

    def __getitem__(self, key: str) -> List[str]:
        return super().__getitem__(key)

    def __str__(self) -> str:
        result = ""
        for query, items in self.items():
            for response in items:
                result += "\n- " + self.postprocess(f"{query} : {response}")
        result = result.strip()
        return result

    def __repr__(self) -> str:
        return str(self)



from .search import SearchDuckDuckGo, SearchLocalDatabase, SearchReranker, SearchRerankerItem, SearchInterface
from .language_model import LanguageModelGPT3, LanguageModelInterface, Completion
from .hints_dict import HintsDict
from .dialogue_list import DialogueList
from .robot import RobotConfig, CharacterConfig, Session, Robot
from .completion_reactions import CompletionReaction, CompletionReactionAnswer, CompletionReactionSearch, CompletionReactionStorySearch, CompletionReactionResult


from typing import List
import requests
import numpy as np
from .language_model import LanguageModelInterface, Completion


_DEFAULT_PARAMS = {
    "model": "text-davinci-003",
    "temperature": 1.0,
    "max_tokens": 150,
    "frequency_penalty": 1.0,
    "presence_penalty": 1.0,
    "top_p": 1,
    "logprobs": 1,
}
_REQUEST_TIMEOUT = 60.0


class LanguageModelGPT3(LanguageModelInterface):
    def __init__(self, key: str, stop: List[str], **kwargs) -> None:
        self.key = key
        self.stop = stop
        self.params = dict(_DEFAULT_PARAMS, **kwargs, stop=stop)

    def complete(self, prompt: str) -> List[Completion]:
        body = dict(self.params, prompt=prompt.strip())
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.key}",
        }
        response = requests.post(
            "https://api.openai.com/v1/completions",
            headers=headers,
            json=body,
            timeout=_REQUEST_TIMEOUT
        )
        data = response.json()
        assert response.status_code == 200, data.get("error", {}).get("message",
            f"GPT3 response status code is {response.status_code}"
        )
        result = []
        for choice in data["choices"]:
            choice_text = choice["text"].strip()
            choice_token_logprobs = np.array(choice["logprobs"]["token_logprobs"], dtype=np.float64)
            choice_token_probs = np.exp(choice_token_logprobs)
            choice_token_probs_normed = choice_token_probs ** (1/len(choice_token_probs))
            choice_proba = choice_token_probs_normed.prod()
            choice_result = Completion(choice_text, choice_proba)
            result.append(choice_result)
        return result



from dataclasses import dataclass
from typing import Union
from ..hints_dict import HintsDict


@dataclass
class CompletionReactionResult:
    text: Union[str, None]
    hints: HintsDict
    repeat: bool


class CompletionReaction:
    def get_prefix(self) -> str:
        return NotImplementedError()

    def process(self, completion: str, hints: HintsDict) -> CompletionReactionResult:
        raise NotImplementedError()



from .completion_reaction import CompletionReaction, CompletionReactionResult
from ..hints_dict import HintsDict


class CompletionReactionAnswer(CompletionReaction):
    def get_prefix(self) -> str:
        return "Answer:"

    def process(self, completion: str, hints: HintsDict) -> CompletionReactionResult:
        return CompletionReactionResult(
            completion,
            hints,
            False
        )



from .search_interface import SearchInterface
from .search_duckduckgo import SearchDuckDuckGo
from .search_local_database import SearchLocalDatabase
from .search_reranker import SearchRerankerItem, SearchReranker



from .completion_reaction import CompletionReaction, CompletionReactionResult
from ..search import SearchInterface
from ..hints_dict import HintsDict


class CompletionReactionSearch(CompletionReaction):
    def __init__(self, search: SearchInterface) -> None:
        super().__init__()
        self.search = search

    def get_prefix(self) -> str:
        return "Search:"

    def get_history_key(self, request: str) -> str:
        return request

    def process(self, completion: str, hints: HintsDict) -> CompletionReactionResult:
        requests = completion.split("\n")
        requests = [request.strip().strip("-") for request in requests]
        requests = [request for request in requests if request]
        requests = [request for request in requests if self.get_history_key(request) not in hints]
        for request in requests:
            hints[self.get_history_key(request)] = self.search.search(request)
        return CompletionReactionResult(
            None,
            hints,
            True
        )


class CompletionReactionStorySearch(CompletionReactionSearch):
    def get_prefix(self) -> str:
        return "Story Search:"

    def get_history_key(self, request: str) -> str:
        return f"(Story) {request}"



from .completion_reaction import CompletionReaction, CompletionReactionResult
from .completion_reaction_search import CompletionReactionSearch, CompletionReactionStorySearch
from .completion_reaction_answer import CompletionReactionAnswer



from typing import List
from dataclasses import dataclass


@dataclass
class Completion:
    text: str
    score: float


class LanguageModelInterface:
    def complete(self, prompt: str) -> List[Completion]:
        raise NotImplementedError()



from typing import List
from dataclasses import dataclass
from .search_interface import SearchInterface
from sentence_transformers import CrossEncoder


@dataclass
class SearchRerankerItem:
    search: SearchInterface
    updateable: bool


class SearchReranker(SearchInterface):
    def __init__(self, model: CrossEncoder, search_systems: List[SearchRerankerItem], batch_size: int, top_n: int, score_threshold: float = 0.0) -> None:
        self.model = model
        self.search_systems = search_systems
        self.batch_size = batch_size
        self.top_n = top_n
        self.score_threshold = score_threshold
    
    def add_documents(self, documents: List[str]) -> None:
        for system in self.search_systems:
            if system.updateable:
                system.search.add_documents(documents)
    
    def search(self, query: str) -> List[str]:
        options = []
        for system in self.search_systems:
            options += system.search.search(query)
        pairs = [
            (query, text)
            for text in options
        ]
        scores = self.model.predict(pairs, batch_size=self.batch_size)
        top_indices = (-scores).argsort()[:self.top_n]
        return [
            options[index]
            for index in top_indices
            if scores[index] >= self.score_threshold
        ]
